############################
# IMPORT
############################

import pygame

############################
# SCENE
############################

def loadTileMap(image: pygame.surface.Surface, tileSize: pygame.Vector2) -> list[pygame.surface.Surface]:
	if image.get_width() % tileSize.x != 0:
		raise Exception("The width of the provided image dose not match the tile width")
	elif image.get_height() % tileSize.y != 0:
		raise Exception("The height of the provided image dose not match the tile height")

	tiles: list[pygame.surface.Surface] = []

	for y in range(int(image.get_height() / tileSize.y)):
		for x in range(int(image.get_width() / tileSize.x)):
			tiles.append(image.subsurface(pygame.rect.Rect(x * tileSize.x, y * tileSize.y, tileSize.x, tileSize.y)))

	return tiles


############################
# IMPORT
############################

import pygame

############################
# OBJECT
############################

class Object:
	def __init__(self) -> None:
		self.init()

	def init(self) -> None:
		pass

	def update(self, events: list[pygame.event.Event], dt: float) -> None:
		pass

	def draw(self, surface: pygame.surface.Surface) -> None:
		pass


############################
# IMPORT
############################

import json
import os
import time
import dataclasses
from datetime import datetime
import glob

############################
# SAVELOAD
############################

@dataclasses.dataclass()
class Save:
    created: datetime
    updated: datetime
    version: int
    data: dict
    

def save(fileName: str, data: dict):
    if not os.path.isdir("saves"):
        os.mkdir("saves")
    
    dataFormated = {
        "createdTime": time.time(),
        "updatedTime": time.time(),
        "version": 1,
        "data": data
    }


    if os.path.exists(fileName + ".save"):
        with open(fileName + ".save", "r") as f:
            dataFormated = json.load(f)

        dataFormated["data"] = data
        dataFormated["version"] = dataFormated["version"] + 1
        dataFormated["updatedTime"] = time.time()

    with open(fileName + ".save", "w") as f:
        f.write(json.dumps(dataFormated))

def load(fileName: str) -> Save:
    data = None

    with open(fileName + ".save", "r") as f:
        data = json.load(f)

    return Save(
        data["createdTime"],
        data["updatedTime"],
        data["version"],
        data["data"]
    )

def findSaves(dir: str) -> list[str]:
    if not os.path.isdir(dir):
        return []

    return glob.glob(dir + "/*.save")



############################
# IMPORT
############################

from typing import Callable
import pygame
from pygame import Vector2
from datetime import datetime
import traceback
from .scene import SceneManager
from .debuginfo import DebugInfo
from dataclasses import fields

############################
# GAME
############################


class Game:
    def __init__(
            self,
            productionMode: bool,
            gameName: str,
            screenSize: Vector2 = Vector2(100, 100),
            canvasSize: Vector2 | None = None,
            screenInitMethod: Callable[[], pygame.surface.Surface] | None = None,
            fps: int = 30):
        self.display: pygame.surface.Surface

        if screenInitMethod:
            self.display = screenInitMethod()
        else:
            self.display = pygame.display.set_mode(screenSize)

        pygame.init()

        # Vars
        self.run: bool = True
        self.clock: pygame.time.Clock = pygame.time.Clock()
        self.fps: int = fps
        self.canvasSize: Vector2 = canvasSize if canvasSize else screenSize
        self.productionMode: bool = productionMode
        self.gameName = gameName
        self.sceneManager = SceneManager()
        self.showDebugInfo: bool = False

        if not self.productionMode:
            self.debugInfo: DebugInfo = DebugInfo()
            self.debugFont = pygame.font.SysFont("arial", 24)

        # Funcs
        if self.productionMode:
            try:
                self.init()
                self.loop()
            except Exception as err:
                self._error(err)
        else:
            print("Use [F1] to show/hide the debug info menu")

            self.init()
            self.loop()

    def _error(self, err: Exception):
        pygame.quit()
        self.run = False

        fileName: str = f"error-log-{datetime.now().strftime('%m-%d-%-y_%h-%m-%s')}.txt"

        with open(fileName, "w") as f:
            f.write(traceback.format_exc())

    def _drawDebugInfo(self, surface: pygame.surface.Surface):
        texts = [f"{field.name}: {getattr(self.debugInfo, field.name)}" for field in fields(self.debugInfo)]

        for index, text in enumerate(texts):
            textSuface = self.debugFont.render(text, False, (0, 0, 0))
            surface.blit(textSuface, Vector2(10, 10 + index * 30))

    def init(self):
        pass

    def loop(self):
        while self.run:
            dt: float = self.clock.tick(self.fps) / 1000

            # Update debuger info data
            if not self.productionMode:
                self.debugInfo.dt = dt
                self.debugInfo.fps = self.clock.get_fps()
                self.debugInfo.scenes = len(self.sceneManager.scenes)
                self.debugInfo.activeScene = self.sceneManager.currentScene.name \
                    if self.sceneManager.currentScene else "Default"

            events = pygame.event.get()

            for event in events:
                if event.type == pygame.KEYDOWN and event.key == pygame.K_F1:
                    self.showDebugInfo = not self.showDebugInfo
                    break

            self.updateMain(events, dt)

            self.sceneManager.update(events, dt)
            self.update(events, dt)

            canvasSize = self.canvasSize if self.canvasSize else self.display.get_size()
            canvas = pygame.surface.Surface(canvasSize)
            if self.sceneManager.hasActiveScene():
                self.sceneManager.draw(canvas)
            else:
                self.draw(canvas)

            if canvasSize != self.display.get_size():
                canvas = pygame.transform.scale(canvas, self.display.get_size())

            if self.showDebugInfo:
                self._drawDebugInfo(canvas)

            self.display.blit(canvas, Vector2(0, 0))

            pygame.display.update()

    def updateMain(self, events: list[pygame.event.Event], dt: float):
        pass

    def update(self, events: list[pygame.event.Event], dt: float):
        pass

    def draw(self, surface: pygame.surface.Surface):
        pass



############################
# IMPORT
############################

import pygame

############################
# SCENE
############################


class Scene:
    def __init__(self):
        self.name = self.__class__.__name__

        self.init()

    def init(self):
        pass

    def update(self, events: list[pygame.event.Event], dt: float):
        pass

    def draw(self, surface: pygame.surface.Surface):
        pass

############################
# SCENE MANAGER
############################


class SceneManager:
    def __init__(self):
        self.scenes: list[Scene | None]= []
        self._currentScene: Scene | None = None

    @property
    def currentScene(self) -> Scene | None:
        return self._currentScene

    @currentScene.setter
    def currentScene(self, value):
        self.scenes.append(self._currentScene)

        for index, scene in enumerate(self.scenes):
            if value == scene:
                self.scenes.pop(index)
                break

        self._currentScene = value

    def removeCurrentScene(self):
        self.currentScene = None

    def hasActiveScene(self) -> bool:
        return self.currentScene is not None

    def update(self, events: list[pygame.event.Event], dt: float):
        if self._currentScene:
            self._currentScene.update(events, dt)

    def draw(self, surface: pygame.surface.Surface):
        if self._currentScene:
            self._currentScene.draw(surface)
